from torch import Tensor

from quaterion.loss.similarity_loss import SimilarityLoss


class PairwiseLoss(SimilarityLoss):
    def __init__(self, distance_metric_name: str = "cosine_distance"):
        super(PairwiseLoss, self).__init__(distance_metric_name=distance_metric_name)

    def forward(
        self, embeddings: Tensor, pairs: Tensor, labels: Tensor, subgroups: Tensor
    ) -> Tensor:
        """
        :param embeddings: shape: [batch_size x vector_length]
        :param pairs: shape: [pairs_count x 2] - contains a list of known similarity pairs in batch
        :param labels: shape: [pairs_count] - similarity of the pair
        :param subgroups: shape: [pairs_count x 2] - subgroup ids of objects
        :return:
        """
        raise NotImplementedError()



from quaterion.dataset.cache_data_loader import CacheDataLoader
from quaterion.dataset.similarity_data_loader import (
    GroupSimilarityDataLoader,
    PairsSimilarityDataLoader,
    SimilarityDataLoader,
)
from quaterion.dataset.similarity_samples import (
    SimilarityGroupSample,
    SimilarityPairSample,
)



from typing import Collection, Hashable, Any, List, Callable, Union, Tuple

from torch import Tensor

from quaterion_models.encoders import Encoder
from quaterion_models.types import TensorInterchange

KeyExtractorType = Callable[[Any], Hashable]

CacheCollateFnType = Callable[
    [Collection[Any]],
    Union[List[Hashable], Tuple[Hashable, TensorInterchange]],
]


class CacheEncoder(Encoder):
    def __init__(self, encoder: Encoder, key_extractor: KeyExtractorType = None):
        if encoder.trainable():
            raise ValueError("Trainable encoder can't be cached")
        super().__init__()
        self._encoder = encoder
        self.cache_filled = False
        self.key_extractor = (
            key_extractor if key_extractor is not None else self.default_key_extractor
        )

    def trainable(self) -> bool:
        return False

    def embedding_size(self) -> int:
        """
        :return: Size of resulting embedding
        """
        return self._encoder.embedding_size()

    @classmethod
    def default_key_extractor(cls, obj: Any) -> Hashable:
        """
        Generate hashable from batch object
        :return: Key for cache
        """
        return (
            hash(obj) if not isinstance(obj, dict) else hash(tuple(sorted(obj.items())))
        )

    def key_collate_fn(self, batch: Collection[Any]) -> List[Hashable]:
        """
        Convert batch of raw data into list of keys for cache
        :return: List of keys for cache
        """
        return [self.key_extractor(value) for value in batch]

    def cache_collate(
        self, batch: Collection[Any]
    ) -> Union[List[Hashable], Tuple[Hashable, TensorInterchange]]:
        """
        Converts raw data batch into suitable model input and keys for caching

        :return: Tuple of keys and according model input
        """
        keys: List[Hashable] = self.key_collate_fn(batch)
        if self.cache_filled:
            return keys
        values: TensorInterchange = self._encoder.get_collate_fn()(batch)
        return keys, values

    def get_collate_fn(self) -> CacheCollateFnType:
        """
        Provides function that converts raw data batch into suitable model
        input

        :return: Model input
        """
        return self.cache_collate

    def forward(self, batch: TensorInterchange) -> Tensor:
        """
        Infer encoder - convert input batch to embeddings

        :param batch: processed batch
        :return: embeddings, shape: [batch_size x embedding_size]
        """
        raise NotImplementedError()

    def save(self, output_path: str):
        """
        Persist current state to the provided directory

        :param output_path:
        :return:
        """
        self._encoder.save(output_path)

    @classmethod
    def load(cls, input_path: str) -> "Encoder":
        """
        CachedEncoder classes wrap already instantiated encoders and don't
        provide loading support.

        :param input_path:
        :return:
        """
        raise ValueError("Cached encoder does not support loading")

    def fill_cache(
        self,
        data: Collection[Hashable],
    ):
        """
        Applies encoder to data and store results in cache

        :param data: collection of hashables to which encoder will be applied
        and the resulting embeddings will be stored in cache
        """
        raise NotImplementedError()

    def reset_cache(self):
        """
        Reset all stored data
        """
        raise NotImplementedError()



from torch import Tensor, LongTensor

from quaterion.loss.similarity_loss import SimilarityLoss


class GroupLoss(SimilarityLoss):
    def __init__(self, distance_metric_name: str = "cosine_distance"):
        super(GroupLoss, self).__init__(distance_metric_name=distance_metric_name)

    def forward(self, embeddings: Tensor, groups: LongTensor) -> Tensor:
        """
        :param embeddings: shape: [batch_size x vector_length]
        :param groups: shape: [batch_size] - Groups, associated with embeddings
        :return: 0-size tensor
        """
        raise NotImplementedError()



from typing import Tuple, Hashable, Iterable

import torch

from torch import Tensor
from quaterion_models.encoders import Encoder
from quaterion_models.types import TensorInterchange

from quaterion.train.encoders.cache_config import CacheType
from quaterion.train.encoders.cache_encoder import (
    CacheEncoder,
    CacheCollateFnType,
    KeyExtractorType,
)


class InMemoryCacheEncoder(CacheEncoder):
    """CacheEncoder which is able to store tensors on CPU or GPU"""

    def __init__(
        self,
        encoder: Encoder,
        key_extractor: KeyExtractorType = None,
        cache_type=CacheType.AUTO,
    ):
        super().__init__(encoder, key_extractor)
        self.cache = {}
        self._cache_type = self.resolve_cache_type(cache_type)

    @staticmethod
    def resolve_cache_type(cache_type: CacheType) -> CacheType:
        """Resolve received cache type.

        If cache type is AUTO, then set it cuda if it is available, otherwise
        use cpu.

        Args:
            cache_type: cache type to be resolved

        Returns:
            CacheType
        """
        if cache_type == CacheType.AUTO:
            cache_type = CacheType.GPU if torch.cuda.is_available() else CacheType.CPU
        return cache_type

    @property
    def cache_type(self):
        return self._cache_type

    def forward(self, batch: TensorInterchange) -> Tensor:
        """Infer encoder - convert input batch to embeddings

        Args:
            batch: tuple of keys to retrieve values from cache

        Returns:
            Tensor: embeddings of shape [batch_size x embedding_size]
        """
        embeddings = torch.stack([self.cache[value] for value in batch])
        if self.cache_type == CacheType.CPU:
            device = next(self.parameters(), torch.Tensor(0)).device
            embeddings = embeddings.to(device)
        return embeddings

    def get_collate_fn(self) -> CacheCollateFnType:
        """Provides function that converts raw data batch into suitable input.

        Returns:
            CacheCollateFnType: method that converts raw data batch into
                keys and encoder's input
        """
        return self.cache_collate

    def fill_cache(self, data: Tuple[Iterable[Hashable], TensorInterchange]) -> None:
        """Apply wrapped encoder to data and store processed data on
        corresponding device.

        Args:
            data: Tuple of keys and batches suitable for encoder

        Returns:
            None
        """
        keys, batch = data
        embeddings = self._encoder(batch)
        if self.cache_type == CacheType.CPU:
            embeddings = embeddings.to("cpu")
        self.cache.update(dict(zip(keys, embeddings)))

    def reset_cache(self) -> None:
        """Resets cache.

        Returns:
            None
        """
        self.cache.clear()
        self.cache_filled = False



from enum import Enum
from dataclasses import dataclass, field
from typing import Dict, Optional, Callable, Any, Hashable


class CacheType(str, Enum):
    AUTO = "auto"
    CPU = "cpu"
    GPU = "gpu"


@dataclass
class CacheConfig:
    cache_type: Optional[CacheType] = None
    mapping: Dict[str, CacheType] = field(default_factory=dict)
    key_extractors: Dict[str, Callable[[Any], Hashable]] = field(default_factory=dict)
    batch_size: Optional[int] = 32
    num_workers: Optional[int] = None  # if None - inherited from source dl



from typing import Type

from torch import Tensor, relu, LongTensor

from quaterion.loss.metrics import SiameseDistanceMetric
from quaterion.loss.pairwise_loss import PairwiseLoss


class ContrastiveLoss(PairwiseLoss):
    """
    Contrastive loss. Expects as input two texts and a label of either 0 or 1.
    If the label == 1, then the distance between the two embeddings is reduced. If the label == 0,
     then the distance between the embeddings is increased.

    Further information: http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf

    :param distance_metric_name: Name of the function, that returns a distance between two embeddings.
        The class SiameseDistanceMetric contains pre-defined metrics that can be used
    :param margin: Negative samples (label == 0) should have a distance of at least the margin value.
    :param size_average: Average by the size of the mini-batch.

    """

    @classmethod
    def metric_class(cls) -> Type:
        return SiameseDistanceMetric

    def __init__(
        self,
        distance_metric_name: str = "cosine_distance",
        margin: float = 1.0,
        size_average: bool = True,
    ):
        super(ContrastiveLoss, self).__init__(distance_metric_name=distance_metric_name)
        self.margin = margin
        self.size_average = size_average

    def get_config_dict(self):
        return {
            **super().get_config_dict(),
            "margin": self.margin,
            "size_average": self.size_average,
        }

    def forward(self, embeddings: Tensor, pairs: LongTensor, labels: Tensor, **kwargs):
        rep_anchor = embeddings[pairs[:, 0]]
        rep_other = embeddings[pairs[:, 1]]
        distances = self.distance_metric(rep_anchor, rep_other)
        losses = 0.5 * (
            labels.float() * distances.pow(2)
            + (1 - labels).float() * relu(self.margin - distances).pow(2)
        )
        return losses.mean() if self.size_average else losses.sum()



from collections import defaultdict
from typing import List, Any, Dict, Callable
from torch.utils.data.dataloader import T_co
from torch.utils.data import Dataset

from quaterion.dataset.similarity_data_loader import SimilarityDataLoader
from quaterion.train.encoders.cache_encoder import KeyExtractorType, CacheCollateFnType


class CacheDataLoader(SimilarityDataLoader):
    """DataLoader used to provide convenient way of caching.

    Provides a way to avoid repeated calculations in the context of current
    process.
    Fetches unique objects from batch and then store results in local cache for
    further reuse.

    """

    def __init__(
        self,
        key_extractors: Dict[str, KeyExtractorType],
        cached_encoders_collate_fns: Dict[str, CacheCollateFnType],
        unique_objects_extractor: Callable[[List[Any]], List[Any]],
        dataset: Dataset[T_co],
        **kwargs
    ):
        # We use `pl.trainer.predict` to fill cache, but it recreates
        # dataloader instance internally. Trainer does it by collecting current
        # instance's `**kwargs` and creating a brand-new object of dataloader.
        # `**kwargs` contains `collate_fn` key which is duplicated with ours
        # and raises a TypeError: multiple values for keyword argument.
        kwargs.pop("collate_fn", None)
        super().__init__(dataset, collate_fn=self.cache_collate_fn, **kwargs)
        self.unique_objects_extractor = unique_objects_extractor
        self.cached_encoders_collate_fns = cached_encoders_collate_fns
        self.key_extractors = key_extractors
        self.seen_objects = defaultdict(set)

    @classmethod
    def fetch_unique_objects(cls, batch: List[Any]) -> List[Any]:
        """Fetches unique objects across batch to avoid repeated calculations.

        Args:
            batch: batch of raw data

        Returns:
            List[Any]: batch of unique objects
        """
        pass

    def cache_collate_fn(self, batch: List[T_co]):
        """
        Collate used to cache batches without repeated calculations of
        embeddings in current process

        Args:
            batch:

        Returns:
            Example:
            ```
            {
                "encoder_name": __batch_of_encoder__
            }
            ```
        """
        unique_objects = self.unique_objects_extractor(batch)

        encoder_batches = defaultdict(list)
        for obj in unique_objects:
            for encoder_name, key_extractor in self.key_extractors.items():
                obj_key = key_extractor(obj)
                if obj_key not in self.seen_objects[encoder_name]:
                    # to be cached by this specific encoder
                    encoder_batches[encoder_name].append(obj)
                    self.seen_objects[encoder_name].add(obj_key)

        new_batch = {}
        for encoder_name, collate in self.cached_encoders_collate_fns.items():
            encoder_batch = encoder_batches[encoder_name]
            new_batch[encoder_name] = collate(encoder_batch) if encoder_batch else None

        return new_batch



from dataclasses import dataclass
from typing import Any


@dataclass
class SimilarityPairSample:
    """
    Example:
        ```
        data = [
            # First query group (subgroup)
            SimilarityPairSample(obj_a="cheesecake", obj_b="muffins", score=0.9, subgroup=10),
            SimilarityPairSample(obj_a="cheesecake", obj_b="macaroons", score=0.8, subgroup=10),
            SimilarityPairSample(obj_a="cheesecake", obj_b="candies", score=0.7, subgroup=10),
            SimilarityPairSample(obj_a="cheesecake", obj_b="nutella", score=0.6, subgroup=10),

            # Second query group
            SimilarityPairSample(obj_a="lemon", obj_b="lime", score=0.9, subgroup=11),
            SimilarityPairSample(obj_a="lemon", obj_b="orange", score=0.7, subgroup=11),
            SimilarityPairSample(obj_a="lemon", obj_b="grapefruit", score=0.6, subgroup=11),
            SimilarityPairSample(obj_a="lemon", obj_b="mandarin", score=0.6, subgroup=11),
        ]
        ```
    """

    obj_a: Any
    obj_b: Any
    score: float
    # Consider all examples outside this group as negative samples.
    # By default, all samples belong to group 0 - therefore other samples could
    # not be used as negative examples.
    subgroup: int = 0


@dataclass
class SimilarityGroupSample:
    """
    Represent groups of similar objects all of which should match with one-another within the group.

    Example:
        Faces dataset. All pictures of a single person should have single unique group id.
        In this case NN will learn to match all pictures within the group closer to each-other, but
        pictures from different groups - further.

        ```csv
        file_name,group_id
        barak_obama_1.jpg,371
        barak_obama_2.jpg,371
        barak_obama_3.jpg,371
        elon_musk_1.jpg,555
        elon_musk_2.jpg,555
        elon_musk_3.jpg,555
        leonard_nimoy_1.jpg,209
        leonard_nimoy_2.jpg,209
        ```
    """

    obj: Any
    group: int



from typing import Type, Callable

from torch import nn, Tensor

from quaterion.loss.metrics import SiameseDistanceMetric


class SimilarityLoss(nn.Module):
    @classmethod
    def metric_class(cls) -> Type:
        return SiameseDistanceMetric

    @classmethod
    def get_distance_function(
        cls, function_name: str
    ) -> Callable[[Tensor, Tensor], Tensor]:
        for name, foo in cls.metric_class().__dict__.items():
            if name == function_name:
                return foo.__func__

        raise RuntimeError(
            f"Unknown `distance_metric_name` {function_name},"
            f" available metrics: {cls.metric_class().__dict__.keys()}"
        )

    def __init__(self, distance_metric_name: str = "cosine_distance"):
        super(SimilarityLoss, self).__init__()
        self.distance_metric_name = distance_metric_name
        self.distance_metric = self.get_distance_function(self.distance_metric_name)

    def get_config_dict(self):
        return {"distance_metric_name": self.distance_metric_name}



import torch
import torch.nn as nn
import torch.nn.functional as F
from quaterion.loss.group_loss import GroupLoss
from torch import LongTensor, Tensor


class SoftmaxLoss(GroupLoss):
    """
    Regular cross-entropy loss.

    It is designed to work with the base `GroupLoss` class.
    """

    def __init__(self, embedding_size: int, num_groups: int, temperature: float = 0.05):
        """An implementation of softmax with dot product.

        :param embedding_size: Output dimension of the encoder.
        :param num_groups: Number of groups in the dataset.
        :param temperature: Temperature value to divide logits, defaults to 0.05

        """
        super(GroupLoss, self).__init__()

        self.temperature = temperature
        self.kernel = nn.Parameter(torch.FloatTensor(embedding_size, num_groups))
        nn.init.normal_(self.kernel, std=0.01)

    def forward(self, embeddings: Tensor, groups: LongTensor) -> Tensor:
        """
        :param embeddings: Shape: (batch_size, vector_length) - Output embeddings from the encoder.
        :param groups: Shape: (batch_size,) - Group ids, associated with embeddings.
        :return: 0-size tensor
        """
        # Shape: (batch_size, num_groups)
        logits = torch.mm(embeddings, self.kernel) / self.temperature

        return F.cross_entropy(logits, groups)



from typing import Any, List, Generic

import torch
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import T_co

from quaterion.dataset.similarity_samples import (
    SimilarityPairSample,
    SimilarityGroupSample,
)


class SimilarityDataLoader(DataLoader, Generic[T_co]):
    @classmethod
    def fetch_unique_objects(cls, batch: List[Any]) -> List[Any]:
        """
        Fetch unique objects from batch to avoid calculation of embeddings
        to repeated objects
        """
        raise NotImplementedError()


class PairsSimilarityDataLoader(SimilarityDataLoader[SimilarityPairSample]):
    @classmethod
    def collate_fn(cls, batch: List[SimilarityPairSample]):
        features = [record.obj_a for record in batch] + [
            record.obj_b for record in batch
        ]
        labels = {
            "pairs": torch.LongTensor([[i, i + len(batch)] for i in range(len(batch))]),
            "labels": torch.Tensor([record.score for record in batch]),
            "subgroups": torch.Tensor([record.subgroup for record in batch] * 2),
        }
        return features, labels

    @classmethod
    def fetch_unique_objects(cls, batch: List[SimilarityPairSample]):
        unique_objects = []
        for sample in batch:
            if sample.obj_a not in unique_objects:
                unique_objects.append(sample.obj_a)
            if sample.obj_b not in unique_objects:
                unique_objects.append(sample.obj_b)
        return unique_objects


class GroupSimilarityDataLoader(SimilarityDataLoader[SimilarityGroupSample]):
    @classmethod
    def collate_fn(cls, batch: List[SimilarityGroupSample]):
        features = [record.obj for record in batch]
        labels = {"groups": torch.LongTensor([record.group for record in batch])}
        return features, labels

    @classmethod
    def fetch_unique_objects(
        cls, batch: List[SimilarityGroupSample]
    ) -> List[SimilarityGroupSample]:
        unique_objects = []
        for sample in batch:
            if sample.obj not in unique_objects:
                unique_objects.append(sample.obj)
        return unique_objects



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quaterion',
 'quaterion.dataset',
 'quaterion.eval',
 'quaterion.loss',
 'quaterion.train',
 'quaterion.train.encoders',
 'quaterion.utils']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.5.3,<0.6.0',
 'pytorch-lightning>=1.5.8,<2.0.0',
 'quaterion-models>=0.1.3',
 'torch>=1.8.2']

setup_kwargs = {
    'name': 'quaterion',
    'version': '0.1.4',
    'description': 'Metric Learning fine-tuning framework',
    'long_description': "# Quaterion\n\n>  A dwarf on a giant's shoulders sees farther of the two \n\nA tool collection and framework for fine-tuning metric learning model. \n",
    'author': 'generall',
    'author_email': 'andrey@vasnetsov.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/qdrant/quaterion',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from functools import partial
from typing import Optional, Callable, List, Any

import pytorch_lightning as pl

from torch.utils.data import DataLoader

from quaterion.dataset.similarity_data_loader import (
    PairsSimilarityDataLoader,
    GroupSimilarityDataLoader,
)
from quaterion.loss import GroupLoss, PairwiseLoss
from quaterion.train.trainable_model import TrainableModel


class Quaterion:
    @classmethod
    def combiner_collate_fn(
        cls,
        batch: List[Any],
        features_collate: Callable,
        labels_collate: Callable,
    ):
        features, labels = labels_collate(batch)
        return features_collate(features), labels

    @classmethod
    def fit(
        cls,
        trainable_model: TrainableModel,
        trainer: pl.Trainer,
        train_dataloader: DataLoader,
        val_dataloader: Optional[DataLoader] = None,
    ):

        if isinstance(train_dataloader, PairsSimilarityDataLoader):
            if isinstance(trainable_model.loss, PairwiseLoss):
                train_dataloader.collate_fn = partial(
                    cls.combiner_collate_fn,
                    features_collate=trainable_model.model.get_collate_fn(),
                    labels_collate=train_dataloader.__class__.collate_fn,
                )
            elif isinstance(trainable_model.loss, GroupLoss):
                raise NotImplementedError(
                    "Can't use GroupLoss with PairsSimilarityDataLoader"
                )

        if isinstance(train_dataloader, GroupSimilarityDataLoader):
            if isinstance(trainable_model.loss, GroupLoss):
                train_dataloader.collate_fn = partial(
                    cls.combiner_collate_fn,
                    features_collate=trainable_model.model.get_collate_fn(),
                    labels_collate=train_dataloader.__class__.collate_fn,
                )
            elif isinstance(trainable_model.loss, PairwiseLoss):
                raise NotImplementedError(
                    "Pair samplers are not implemented yet. "
                    "Try other loss/data loader"
                )

        if val_dataloader is not None:
            val_dataloader.collate_fn = train_dataloader.collate_fn

        trainable_model.cache(
            trainer,
            trainable_model.model.encoders,
            train_dataloader,
            val_dataloader,
            trainable_model.cache_config,
        )

        trainer.fit(
            model=trainable_model,
            train_dataloaders=train_dataloader,
            val_dataloaders=val_dataloader,
        )



from quaterion.loss.arcface_loss import ArcfaceLoss
from quaterion.loss.contrastive_loss import ContrastiveLoss
from quaterion.loss.group_loss import GroupLoss
from quaterion.loss.metrics import SiameseDistanceMetric
from quaterion.loss.pairwise_loss import PairwiseLoss
from quaterion.loss.similarity_loss import SimilarityLoss
from quaterion.loss.softmax_loss import SoftmaxLoss



import torch
import torch.nn as nn
import torch.nn.functional as F
from quaterion.loss.group_loss import GroupLoss


def l2_norm(inputs: torch.Tensor, dim: int = 0):
    """
    :param inputs: Input tensor.
    :param dim: Dimension to operate on.
    :return: L2-normalized tensor.
    """
    outputs = inputs / torch.norm(inputs, 2, dim, True)

    return outputs


class ArcfaceLoss(GroupLoss):
    """
    Additive Angular Margin Loss as defined in https://arxiv.org/abs/1801.07698
    """

    def __init__(
        self, embedding_size: int, num_groups: int, s: float = 64.0, m: float = 0.5
    ):
        """
        :param embedding_size: Output dimension of the encoder.
        :param num_groups: Number of groups in the dataset.
        :param s: Scaling value to make cross entropy work.
        :param m: Margin value to push groups apart.

        """
        super(GroupLoss, self).__init__()

        self.kernel = nn.Parameter(torch.FloatTensor(embedding_size, num_groups))
        nn.init.normal_(self.kernel, std=0.01)
        self.s = s
        self.m = m

    def forward(self, embeddings, groups):
        """
        :param embeddings: Shape: (batch_size, vector_length) - Output embeddings from the encoder.
        :param groups: Shape: (batch_size,) - Group ids associated with embeddings.
        :return: Zero-size tensor.
        """
        embeddings = l2_norm(embeddings, 1)
        kernel_norm = l2_norm(self.kernel, 0)

        # Shape: (batch_size, num_groups)
        cos_theta = torch.mm(embeddings, kernel_norm)
        # insure numerical stability
        cos_theta = cos_theta.clamp(-1, 1)

        # Shape: (batch_size,)
        index = torch.where(groups != -1)[0]

        # Shape: (batch_size, num_groups)
        m_hot = torch.zeros(
            index.size()[0], cos_theta.size()[1], device=cos_theta.device
        )
        m_hot.scatter_(1, groups[index, None], self.m)
        cos_theta.acos_()
        cos_theta[index] += m_hot
        cos_theta.cos_().mul_(self.s)

        # calculate scalar loss
        loss = F.cross_entropy(cos_theta, groups)

        return loss



import torch
from torch import Tensor, pairwise_distance, cosine_similarity


class SiameseDistanceMetric:
    """
    The metric for the contrastive loss
    """

    @staticmethod
    def euclidean(x: Tensor, y: Tensor) -> Tensor:
        return pairwise_distance(x, y, p=2)

    @staticmethod
    def manhattan(x: Tensor, y: Tensor) -> Tensor:
        return pairwise_distance(x, y, p=1)

    @staticmethod
    def cosine_distance(x: Tensor, y: Tensor) -> Tensor:
        return 1 - cosine_similarity(x, y)

    @staticmethod
    def dot_product_distance(x: Tensor, y: Tensor) -> Tensor:
        return -torch.einsum("id,jd->j", x, y)



from .cache_encoder import CacheEncoder
from .in_memory_cache_encoder import InMemoryCacheEncoder
from .cache_config import CacheConfig, CacheType

__all__ = [
    "CacheConfig",
    "CacheType",
    "CacheEncoder",
    "InMemoryCacheEncoder",
]



from typing import Dict, Any, Union, Optional

import torch
import pytorch_lightning as pl
from pytorch_lightning.utilities.types import (
    TRAIN_DATALOADERS,
    EVAL_DATALOADERS,
)

from quaterion_models.encoders import Encoder
from quaterion_models.heads import EncoderHead
from quaterion_models import MetricModel
from quaterion.train.encoders import (
    CacheConfig,
    CacheType,
)
from quaterion.loss import SimilarityLoss
from quaterion.utils.enums import TrainStage
from quaterion.train.cache_mixin import CacheMixin


class TrainableModel(pl.LightningModule, CacheMixin):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        encoders = self.configure_encoders()
        self.cache_config = self.configure_caches()
        encoders = self._apply_cache_config(encoders, self.cache_config)

        head = self.configure_head(MetricModel.get_encoders_output_size(encoders))

        self._model = MetricModel(encoders=encoders, head=head)
        self._loss = self.configure_loss()

    @property
    def model(self) -> MetricModel:
        return self._model

    @property
    def loss(self) -> SimilarityLoss:
        return self._loss

    def configure_loss(self) -> SimilarityLoss:
        raise NotImplementedError()

    def configure_encoders(self) -> Union[Encoder, Dict[str, Encoder]]:
        """
        Use this function to define an initial state of encoders.
        This function should be used to assign initial values for encoders
        before training as well as during the checkpoint loading.

        :return: Instance of the `Encoder` or dict of instances
        """
        raise NotImplementedError()

    def configure_caches(self) -> Optional[CacheConfig]:
        """
                Use this method to define which encoders should cache calculated
                embeddings and what kind of cache they should use.

                Examples:

                >>> CacheConfig(CacheType.AUTO)
                CacheConfig(
                    cache_type=<CacheType.AUTO: 'auto'>,
                    mapping={},
                    key_extractors={}
                )

                >>> cache_config = CacheConfig(
        ...     mapping={"text_encoder": CacheType.GPU, "image_encoder": CacheType.CPU}
        ... )
                CacheConfig(
                    cache_type=None,
                    mapping={
                        'text_encoder': <CacheType.GPU: 'gpu'>,
                        'image_encoder': <CacheType.CPU: 'cpu'>
                    },
                    key_extractors={}
                )
                >>> CacheConfig(
        ...     cache_type=CacheType.AUTO,
        ...     key_extractors={"default": lambda obj: hash(obj)}
        ... )
                CacheConfig(
                    cache_type=<CacheType.AUTO: 'auto'>,
                    mapping={},
                    key_extractors={'default': <function <lambda> at 0x106bc90e0>}
                )

                :return: CacheConfig
        """
        pass

    def configure_head(self, input_embedding_size: int) -> EncoderHead:
        """
        Use this function to define an initial state for head layer of the model

        :param input_embedding_size: size of embeddings produced by encoders
        :return: Instance of `EncoderHead`
        """
        raise NotImplementedError()

    def process_results(
        self,
        embeddings: torch.Tensor,
        targets: Dict[str, Any],
        batch_idx,
        stage: TrainStage,
        **kwargs,
    ):
        """
        Define any additional evaluations of embeddings here.

        :param embeddings: Tensor of batch embeddings, shape: [batch_size x embedding_size]
        :param targets: Output of batch target collate
        :param batch_idx: ID of the processing batch
        :param stage: Train, validation or test stage
        :return: None
        """
        pass

    def training_step(self, batch, batch_idx, **kwargs) -> torch.Tensor:
        stage = TrainStage.TRAIN
        loss = self._common_step(
            batch=batch, batch_idx=batch_idx, stage=stage, **kwargs
        )
        return loss

    def validation_step(self, batch, batch_idx, **kwargs) -> Optional[torch.Tensor]:
        stage = TrainStage.VALIDATION
        self._common_step(batch=batch, batch_idx=batch_idx, stage=stage, **kwargs)
        return None

    def test_step(self, batch, batch_idx, **kwargs) -> Optional[torch.Tensor]:
        stage = TrainStage.TEST
        self._common_step(batch=batch, batch_idx=batch_idx, stage=stage, **kwargs)
        return None

    def _common_step(self, batch, batch_idx, stage: TrainStage, **kwargs):
        features, targets = batch
        embeddings = self.model(features)
        loss = self.loss(embeddings=embeddings, **targets)
        self.log(f"{stage}_loss", loss)
        self.process_results(
            embeddings=embeddings,
            targets=targets,
            batch_idx=batch_idx,
            stage=stage,
            **kwargs,
        )
        return loss

    def save_servable(self, path: str):
        """
        Save model, independent of Pytorch Lightning.

        :param path: where to save
        :return: None
        """
        self.model.save(path)

    # region anchors
    # https://github.com/PyTorchLightning/pytorch-lightning/issues/10667
    def train_dataloader(self) -> TRAIN_DATALOADERS:
        pass

    def test_dataloader(self) -> EVAL_DATALOADERS:
        pass

    def val_dataloader(self) -> EVAL_DATALOADERS:
        pass

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        pass

    # endregion



import os
import multiprocessing as mp

from typing import (
    Union,
    Dict,
    Optional,
    Set,
    Any,
    Callable,
    Hashable,
    List,
    Iterable,
    Tuple,
)

import torch.cuda
import pytorch_lightning as pl

from loguru import logger
from torch.utils.data import DataLoader
from pytorch_lightning.loops import (
    FitLoop,
    TrainingEpochLoop,
    TrainingBatchLoop,
    EvaluationLoop,
)
from pytorch_lightning.utilities.types import (
    TRAIN_DATALOADERS,
    EVAL_DATALOADERS,
)
from quaterion_models.encoders import Encoder
from quaterion_models.model import DEFAULT_ENCODER_KEY
from quaterion_models.types import TensorInterchange

from quaterion.dataset.cache_data_loader import CacheDataLoader
from quaterion.train.encoders.cache_encoder import KeyExtractorType
from quaterion.dataset.similarity_data_loader import SimilarityDataLoader
from quaterion.train.encoders import (
    CacheConfig,
    CacheEncoder,
    CacheType,
    InMemoryCacheEncoder,
)


class CacheMixin:
    CACHE_MULTIPROCESSING_CONTEXT = "fork"

    @classmethod
    def _apply_cache_config(
        cls,
        encoders: Union[Encoder, Dict[str, Encoder]],
        cache_config: Optional[CacheConfig],
    ) -> Union[Encoder, Dict[str, Encoder]]:
        """Applies received cache configuration for cached encoders, remain
        non-cached encoders as is

        Args:
            encoders: all model's encoders
            cache_config: CacheConfig instance defined in `configure_cache`
                method of the model

        Returns:
            Union[Encoder, Dict[str, encoder]]: encoder or dict of encoders
                which were wrapped into CacheEncoder instances according to
                received cache config.
                Result type depends on the way encoder was defined in the
                model: with or without explicit mapping

        Raises:
            KeyError: encoder's name in cache config is not in model's
                encoders
            ValueError: if CacheConfig instance does not have some of required
                options set. E.g. not `mapping` nor `cache_type` being set
        """
        if not cache_config:
            return encoders

        if cache_config.mapping:
            if cache_config.cache_type:
                logger.warning(
                    "CacheConfig.cache_type has no effect when mapping is set"
                )

            possible_cache_encoders: Set[str] = {
                encoder_name
                for encoder_name in encoders
                if not encoders[encoder_name].trainable()
            }

            for encoder_name, cache_type in cache_config.mapping.items():
                encoder: Optional[Encoder] = encoders.get(encoder_name)
                if not encoder:
                    raise KeyError(
                        f"Can't configure cache for encoder {encoder_name}. "
                        "Encoder not found"
                    )
                cls._check_cuda(cache_type, encoder_name)
                key_extractor: Optional[
                    Callable[[Any], Hashable]
                ] = cache_config.key_extractors.get(encoder_name)

                encoders[encoder_name]: CacheEncoder = cls._wrap_encoder(
                    encoder,
                    cache_type,
                    key_extractor,
                    encoder_name,
                )

                possible_cache_encoders.remove(encoder_name)

            not_cached_encoders = ", ".join(possible_cache_encoders)
            if not_cached_encoders:
                logger.info(
                    f"{not_cached_encoders} haven't been cached, "
                    "but could be as non-trainable encoders"
                )

        elif cache_config.cache_type:
            encoder_name = DEFAULT_ENCODER_KEY

            cls._check_cuda(cache_config.cache_type, encoder_name)
            key_extractor = cache_config.key_extractors.get(encoder_name)
            encoders = cls._wrap_encoder(
                encoders,
                cache_config.cache_type,
                key_extractor,
                encoder_name,
            )
        else:
            raise ValueError(
                "If cache is configured, cache_type or mapping have to be set"
            )

        return encoders

    @staticmethod
    def _check_cuda(cache_type: CacheType, encoder_name: str) -> None:
        """
        If encoder is supposed to use GPU as tensor's storage, then cuda
        has to be available, otherwise raises ValueError

        Args:
            cache_type: cache type for encoder
            encoder_name: name of cache encoder

        Returns:
            None

        Raises:
            ValueError: If cuda is not available
        """
        if cache_type == CacheType.GPU and not torch.cuda.is_available():
            raise ValueError(
                f"`CacheType.GPU` has been chosen for `{encoder_name}` "
                "encoder, but cuda is not available"
            )

    @staticmethod
    def _wrap_encoder(
        encoder: Encoder,
        cache_type: CacheType,
        key_extractor: Optional[KeyExtractorType],
        encoder_name: str = "",
    ) -> CacheEncoder:
        """Wrap encoder into CacheEncoder instance.

        Args:
            encoder: raw model's encoder
            cache_type: cache type of tensor storage
            key_extractor: function to obtain hashable values for complex
                object which can't be hashed with standard means
            encoder_name: name of encoder to be wrapped

        Returns:
            CacheEncoder: wrapped encoder

        Raises:
            ValueError: if encoder layers are not frozen. Cache can be applied
                only to fully frozen encoders' outputs.
        """
        if encoder.trainable():
            raise ValueError(
                f"Can't configure cache for encoder {encoder_name}. "
                "Encoder must be frozen to cache it"
            )

        encoder = InMemoryCacheEncoder(encoder, key_extractor, cache_type)

        return encoder

    @classmethod
    def cache(
        cls,
        trainer: pl.Trainer,
        encoders: Dict[str, Encoder],
        train_dataloader: SimilarityDataLoader,
        val_dataloader: Optional[SimilarityDataLoader],
        cache_config: CacheConfig,
    ) -> None:
        """Filling cache for model's cache encoders.

        Args:
            trainer: performs one training epoch to cache encoders outputs.
                Preserve all encapsulated pytorch-lightning logic such as
                device managing etc.
            encoders: mapping of model's encoders and their names
            train_dataloader: model's train dataloader
            val_dataloader: model's val dataloader
            cache_config: cache config instance to configure cache batch size
                and num of workers to use for caching

        Returns:
            None
        """
        cache_encoders = {
            name: encoder
            for name, encoder in encoders.items()
            if isinstance(encoder, CacheEncoder)
        }

        if not cache_encoders:
            return

        cls._compatibility_check(train_dataloader)
        train_dataloader = cls._wrap_cache_dataloader(
            train_dataloader, cache_config, cache_encoders
        )

        if val_dataloader is not None:
            val_dataloader = cls._wrap_cache_dataloader(
                val_dataloader, cache_config, cache_encoders
            )

        cls._fill_cache(trainer, cache_encoders, train_dataloader, val_dataloader)

        # Once cache is filled, collate functions return only keys for cache
        for encoder in cache_encoders.values():
            encoder.cache_filled = True
        logger.info("Caching has been successfully finished")

    @classmethod
    def _fill_cache(
        cls,
        trainer: pl.Trainer,
        cache_encoders: Dict[str, CacheEncoder],
        train_dataloader: SimilarityDataLoader,
        val_dataloader: SimilarityDataLoader,
    ) -> None:
        """Fills cache and restores trainer state for further training process.

        Args:
            trainer: performs one training and validation epoch
            cache_encoders: mapping of encoders to cache input
            train_dataloader: model's train dataloader
            val_dataloader: model's val dataloader

        Returns:
            None
        """
        # store configured fit and validate loops to restore them for training
        # process after cache
        _fit_loop = trainer.fit_loop
        _val_loop = trainer.validate_loop

        # Mimic fit loop configuration from trainer
        fit_loop = FitLoop(
            min_epochs=1,
            max_epochs=1,
        )
        training_epoch_loop = TrainingEpochLoop()
        training_batch_loop = TrainingBatchLoop()
        training_validation_loop = EvaluationLoop()
        training_epoch_loop.connect(
            batch_loop=training_batch_loop, val_loop=training_validation_loop
        )
        fit_loop.connect(epoch_loop=training_epoch_loop)
        trainer.fit_loop = fit_loop

        # The actual caching
        trainer.predict(
            CacheModel(
                cache_encoders,
            ),
            [train_dataloader, val_dataloader],
        )
        trainer.fit_loop = _fit_loop

    @classmethod
    def _compatibility_check(cls, dataloader: DataLoader) -> None:
        """Checks whether dataloader type.

        To be used in caching, dataloader has to have an implementation of
        `cache_collate_fn`. Currently, only `SimilarityDataLoader` has it.

        Args:
            dataloader: DataLoader

        Returns:
            None

        Raises:
            TypeError: if dataloader is not instance of SimilarityDataLoader
        """
        if not isinstance(dataloader, SimilarityDataLoader):
            raise TypeError("DataLoader must be SimilarityDataLoader ")

    @classmethod
    def _wrap_cache_dataloader(
        cls,
        dataloader: SimilarityDataLoader,
        cache_config: CacheConfig,
        cache_encoders: Dict[str, CacheEncoder],
    ) -> CacheDataLoader:
        """Wrap dataloader for caching.

        Child processes need to derive randomized `PYTHONHASHSEED` value from
        parent process to obtain the same hash values. It is only done with
        `fork` start method.
        `fork` start method is presented on Unix systems, and it is the default
        for them, except macOS. Therefore, we set `fork` method explicitly.

        If dataloader is not supposed to use child process, nothing being done.
        If multiprocessing_context was set by user, it is being untouched and
        can lead to errors.
        If `PYTHONHASHSEED` is set explicitly then multiprocessing_context
        won't be switched.
        Cache can be used on Windows only in case when `PYTHONHASHSEED` is set
        explicitly.

        Args:
            dataloader: dataloader to be wrapped
            cache_config: cache config to retrieve num of workers and batch
                size
            cache_encoders: encoders to set key extractors and collate_fns

        Returns:
            CacheDataLoader: wrapped dataloader
        """
        cls._switch_multiprocessing_context(dataloader)
        num_workers = (
            cache_config.num_workers
            if cache_config.num_workers is not None
            else dataloader.num_workers
        )

        if num_workers == 0:
            mp_ctx = None
        elif dataloader.multiprocessing_context:  # already switched or
            # set by user
            mp_ctx = dataloader.multiprocessing_context
        elif "PYTHONHASHSEED" in os.environ:  # source dataloader has no
            # mp context set, use default on current OS
            mp_ctx = mp.get_start_method()
        else:
            mp_ctx = cls.CACHE_MULTIPROCESSING_CONTEXT
            cls._check_mp_context(mp_ctx)

        # We need to reduce random sampling and repeated calculations to
        # make cache as fast as possible. Thus, we recreate dataloader
        # and set batch size explicitly.
        cache_dl = CacheDataLoader(
            key_extractors=dict(
                (encoder_name, cache_encoder.key_extractor)
                for encoder_name, cache_encoder in cache_encoders.items()
            ),
            cached_encoders_collate_fns=dict(
                (encoder_name, cache_encoder.get_collate_fn())
                for encoder_name, cache_encoder in cache_encoders.items()
            ),
            unique_objects_extractor=dataloader.fetch_unique_objects,
            dataset=dataloader.dataset,
            batch_size=cache_config.batch_size,
            num_workers=num_workers,
            pin_memory=dataloader.pin_memory,
            timeout=dataloader.timeout,
            worker_init_fn=dataloader.worker_init_fn,
            prefetch_factor=dataloader.prefetch_factor,
            multiprocessing_context=mp_ctx,
        )
        return cache_dl

    @classmethod
    def _switch_multiprocessing_context(cls, dataloader: SimilarityDataLoader) -> None:
        """Switch dataloader multiprocessing context.

        Do nothing if dataloader is not supposed to use child processes or
        `PYTHONHASHSEED` has been set explicitly by user.

        Args:
            dataloader: dataloader to check and switch multiprocessing context

        Returns:
            None
        """
        if "PYTHONHASHSEED" in os.environ:
            return

        if dataloader.num_workers == 0:
            return

        mp_context: Optional[
            Union[str, mp.context.BaseContext]
        ] = dataloader.multiprocessing_context
        cls._check_mp_context(mp_context)

        dataloader.multiprocessing_context = cls.CACHE_MULTIPROCESSING_CONTEXT

    @classmethod
    def _check_mp_context(
        cls, mp_context: Optional[Union[str, mp.context.BaseContext]]
    ) -> None:
        """Check if multiprocessing context is compatible with cache.

        Emits warning if current multiprocessing context start method does not
        coincide with one required by cache.

        Args:
            mp_context: some dataloader's multiprocessing context

        Returns:
            None

        Raises:
            OSError: Raise OSError if OS does not support process start method
            required by cache. Currently, this start method is `fork` which is
            not supported by Windows.
        """
        if not mp_context:
            return

        if not isinstance(mp_context, str):
            mp_context = mp_context.get_start_method()

        if mp_context != cls.CACHE_MULTIPROCESSING_CONTEXT:
            logger.warning(
                "Default start method on your OS is not "
                f"{cls.CACHE_MULTIPROCESSING_CONTEXT}. "
                "Trying to switch it. However "
                f"{cls.CACHE_MULTIPROCESSING_CONTEXT} may be unsafe or "
                "unsupported. The most safe option is to launch your process "
                "with fixed `PYTHONHASHSEED` env and remain `spawn` as start "
                "method.\n"
                "Possible launch is `PYTHONHASHSEED=0 python3 run.py"
            )

        if cls.CACHE_MULTIPROCESSING_CONTEXT not in mp.get_all_start_methods():
            raise OSError(
                f"Cache can't be used without setting `PYTHONHASHSEED`. "
                f"{cls.CACHE_MULTIPROCESSING_CONTEXT} multiprocessing context "
                "is not available on current OS"
            )


class CacheModel(pl.LightningModule):
    """Mock model for convenient caching.

    This class is required to make caching process similar to the training of
    the genuine model and inherit and use the same trainer instance. It allows
    avoiding of messing with device managing stuff and more.
    """

    def __init__(
        self,
        encoders: Dict[str, CacheEncoder],
    ):
        super().__init__()
        self.encoders = encoders
        for key, encoder in self.encoders.items():
            self.add_module(key, encoder)

    def predict_step(
        self,
        batch: Dict[str, Tuple[Iterable[Hashable], TensorInterchange]],
        batch_idx: int,
        dataloader_idx: Optional[int] = None,
    ):
        """Caches batch of input.

        Args:
            batch: batch of collated data. Contains mapping, where key is
                encoder's name, value is tuple of key used in cache and
                according items, which are ready to be processed by specific
                encoder.
            batch_idx: Index of current batch
            dataloader_idx: Index of the current dataloader
        Returns:
            torch.Tensor: loss mock
        """
        for encoder_name, encoder in self.encoders.items():
            encoder_samples = batch.get(encoder_name)
            if not encoder_samples:
                continue
            encoder.fill_cache(encoder_samples)

        return torch.Tensor([1])

    # region anchors
    def train_dataloader(self) -> TRAIN_DATALOADERS:
        pass

    def test_dataloader(self) -> EVAL_DATALOADERS:
        pass

    def val_dataloader(self) -> EVAL_DATALOADERS:
        pass

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        pass

    # endregion
